; Formula construction functions, uses definitions in graph.lisp
; Version with pointer structures

(defvar LXL 0 "lexical level")

(defun make-graph (form)
"(make-graph formula)  -- creates graph of (translated) formula, having
                          erased previous graphs."
 (setq LXL 0) (if (plusp ROOT) (-F ROOT))
 (cleanup)
 (setq ROOT (graph (translate form))) (+F ROOT) ROOT
)

(defmacro define (sym form)
"(define 'sym' 'form') -- install (new) notion as sym"
 `(defntn ',sym ',form)
)

(defun defntn (sym form)
"(defntn sym form) -- function for define"
 (cleanup)
 (if (get sym 'OPERATOR) (cons sym '(IS AN OPERATOR - NOT DEFINABLE))
     (let ((s (graph sym)) (e (graph (translate form))) (p (newc)))
       (sG s (ty-notn)) (sR s p) (sG p (ty-point)) (sR p e)
       (sN s 1) (+F s) (+F p) (+F e) s))
)

(defun graph (frm)
 "(graph formula)  -- Constructs graph for formula in environment in CM"
  (prog (node a fl)
  (return
   (cond ((atom frm)
	  (if (numberp frm)
	      (progn (setq node (newc))
                     (sG node (ty-numb)) (sX node frm) (sN node 1) node)
	      (sym-node frm)))
	 ((= (setq fl (length frm)) 2)
	  (setq node (newc))
	  (sG node (ty-comb))
	  (make-left node (graph (car frm)))
	  (make-right node (graph (cadr frm)))
	  node)
	 ((and (= fl 3) (eq (car frm) 'exists))
	  (make-vars (second frm))
	  (setq node (graph (third frm)))
	  (finish-vars node)
          node)
	 ((and (= fl 3) (eq (car frm) 'cons))
	  (setq node (newc))
	  (sG node (ty-cons))
	  (make-left node (graph (cadr frm)))
	  (make-right node (graph (caddr frm)))
	  node)
	 ((> fl 2)
	  (if (eq (car frm) 'lambda)
	      (progn (format T "untranslated lambda expression~%") 0)
	      (multiple-value-bind (fr ls) (but-last frm)
		 (setq node (newc))
		 (sG node (ty-comb))
		 (make-left node (graph fr))
		 (make-right node (graph ls))
		 node)))
	 )))
)

(defun sym-node (sym)
"(sym-node sym) -- Returns cb-addr of current meaning of symbol sym.
                   Creates symbol rep as needed."
 (prog ((n (get sym 'CMSYM)) a)
       (when (null n) (setq n (entersym sym)) (go NEW))
       (setq a (meaning n))
       (if (plusp a) (return a))
  NEW  (setq a (newa)) (sG a (ty-dict)) (sX a n) (sN a 1) (sS a 1) (sE a T)
       (return a))
	 
)

(defun meaning (n)
"(meaning n) -- returns cube address of current proc with name-code n,
               0 if there is none."
 (*when (and!! ENVT (=!! XNMB (!! n))) (*logior (self-address!!)))
)

(defun make-vars (vars)
"(make-vars vars) --- returns env augmented with new vars
                     Records var symbols."
 (incf LXL)
 (dolist (v vars)
  (let ((n (get v 'CMSYM)) (c (newc)) (e 0))
    (if (null n) (setq n (entersym v)) (setq e (meaning n)))
    (sG c (ty-var)) (sL c LXL) (sX c n) (sF c 1) (sR c e)
    (sE e NIL) (sE c T) (+F e))
 )
)

(defun finish-vars (node)
"(finish-var node) -- finishes processing of vars at current LXL associated
                     with node."
 (*when (and!! (VARP!!) (=!! LPTR (!! LXL)))
    (sC!! (!! node))  (+Q!! CXTP 1)  (sL!! (!! 0))
    (*when (plusp!! RPTR) (*pset :overwrite T!! ENVT RPTR))
    (*when (>!! RPTR (!! syscmx)) (+F!! RPTR -1))
    (sE!! NIL!!)  (sR!! (!! 0))
    (sF!! (1-!! REFC)))
 (decf LXL)
)

(defun make-left (n l)
 "(make-left node ln)  -- make L[node] point to ln"
  (sL n l) (+F l))

(defun make-right (n r)
"(make-right node rn)  -- make R[node] point to rn"
  (sR n r) (+F r))

(defun later () "For items not yet implemented" 0)

(defun newa ()
"(newa)  -- returns cube-address of free atom.  Node remains free.
           NIL if there is none"
(let ((r (*all (*when (and!! (FREEP!!) (<=!! (self-address!!) (!! CCPL)))
		      (*min (self-address!!))))))
  (when r (if (plusp LFREE) (decf LFREE) (incf ATML)))
  r)
)

(defun newc ()
"(newc)  -- returns cube-address of free ccpl.  Node remains free.
           NIL if there is none"
(let ((r (*all (*when (and!! (FREEP!!) (>=!! (self-address!!) (!! ATML)))
		      (*max (self-address!!))))))
  (sC r r)
  (when r (if (plusp HFREE) (decf HFREE) (decf CCPL)))
  r)
)

(defun but-last (lst)
"(but-last list)  --  values (<list except last entry>, <last entry>)
                     list should have 2 or more entries."
 (let (ls)
  (labels ((blst (lst) 
		 (if (null (cdr lst))
		     (progn (setq ls (car lst)) nil)
		     (cons (car lst) (blst (cdr lst))))))
    (values (blst lst) ls)))
)

(defun c-t-abs (v b)
"(c-t-abs var body)  -- returns Curry-Turner abstraction of var from body,
                       a formula, not a graph."
 (let (opr opd abs-opr abs-opd (v-in-opr nil) (v-in-opd nil))
   (when (longexpr b) (multiple-value-setq (opr opd) (but-last b))
		      (when (null (cdr opr)) (setq opr (car opr)))
		      (setq v-in-opr (occurs v opr))
		      (setq v-in-opd (occurs v opd)))
   (cond ((eq v b) '*I)
	 ((not (or v-in-opr v-in-opd )) (list '*K b))
	 ((and (= (length b) 3) (eq (car b) 'CONS))
	  (list 'CONS (c-t-abs v (second b)) (c-t-abs v (third b))))
	 ((eq v opd) (if v-in-opr (list '*W (c-t-abs v opr)) opr))
	 ((and v-in-opr v-in-opd)
	  (if (and (consp (setq abs-opr (c-t-abs v opr)))
		   (= (length abs-opr) 3)
		   (eq '*B (car abs-opr)))
	      (list '*SB (cadr abs-opr) (caddr abs-opr) (c-t-abs v opd))
	      (list '*S  abs-opr (c-t-abs v opd))))
	 ((and v-in-opd (not v-in-opr))
	  (if (and (consp (setq abs-opd (c-t-abs v opd)))
		   (= (length abs-opd) 3)
		   (eq '*B (car abs-opd)))
	      (list '*BB opr (cadr abs-opd) (caddr abs-opd))
	      (list '*B  opr abs-opd)))
	 ((and v-in-opr (not v-in-opd))
	  (if (and (consp (setq abs-opr (c-t-abs v opr)))
		   (= (length abs-opr) 3)
		   (eq '*B (car abs-opr)))
	      (list '*CB (cadr abs-opr) (caddr abs-opr) opd)
	      (if (member abs-opr '(= + * and or)) (list abs-opr opd)
                  (list '*C  abs-opr opd))))))
)


(defun longexpr (xpr)
"longexpr expr  -- expr is a list-structure of length >= 2."
 (and (consp xpr) (consp (cdr xpr))))


(defun occurs (v b)
"occurs var expr  -- var occurs in expression."
 (or (eq v b) (and (consp b) (or (occurs v (car b)) (occurs v (cdr b)))))
)


(defun translate (y)
"(translate expr)  -- returns combinatorial form of expression
                     containing lambdas, equations, assoc forms."
 (let ((x (pre y)))
  (cond ((atom x) x)
        ((and (consp x) (= (length x) 3) (eq (car x) 'lambda))
	 (c-t-abs (cadr x) (translate (caddr x))))
	(T (mapcar #'translate x))))
)

(defvar reportall nil)

(defun report ()
"report  -- prints status of non-free processors"
 (format T
" node  type    L      R      C     Number    Refc   Qntc  ~
N S V E P (Symbol)~%")
 (format T
"------ ---- ------ ------ ------ ---------- ------ ------ ~
- - - - - --------~%")
 (pause NIL) (pause T) (pause T)
 (let ((err nil))
 (*when (and!! (/=!! (GTYP) (!! (ty-free))) (/=!! (GTYP) (!! (ty-dummy)))
	       (or!! (!! reportall) (/=!! (GTYP) (!! (ty-sysc)))))
  (do-for-selected-processors (i)
   (let ((ty (G i))) 
   (when (null err)
    (if (member (pause "<return> to continue, q to quit") '(#\q #\Q))
        (setq err 'STOPPED)
   (progn
    (format T "~6D ~4A ~6D ~6D ~6D ~10D ~6D ~6D ~1A ~1A ~1A ~1A ~1A"
	    i (ex-type ty) (L i) (R i) (C i) (X i) (F i) (Q i)
	    (bval (N i)) (bval (S i)) (bval (V i)) (bval (E i))
	    (bval (pref PCXT i)))
    (if (or (= ty (ty-var)) (= ty (ty-dict)) (= ty (ty-sysc)) (= ty (ty-notn)))
	(format T " ~A~%" (aref symvec (X i)))
	(format T "~%"))))) ; end (when (null err) ... )
    )))  ; end (*when ...)
 (if err err (values)))  ; end (let ((err nil)) ...)
)

(defun bval (x)
"(bval x) -- T/F as x is (non-NIL non-0)/(NIL 0)"
(if (or (null x) (zerop x)) 'F 'T)
)

(defun ex-type (ty)
"ex-type type-code  -- returns symbolic representation of type-code"
      (cond ((= ty (ty-numb))   'numb)
	    ((= ty (ty-comb))   'comb)
	    ((= ty (ty-cons))   'cons)
	    ((= ty (ty-var))    'var)
	    ((= ty (ty-point))  'pntr)
	    ((= ty (ty-dict))   'dict)
	    ((= ty (ty-sysc))   'sysc)
	    ((= ty (ty-notn))   'notn)
	    ((= ty (ty-new))    'new)
	    ((= ty (ty-dead))   'dead)
	    (T 'bad-type))
)

(defvar pausek 31 "Number of lines per page")
(defvar pausec 0)

(defun pause (msg)
"(pause nil) -- initialize line counts
(pause msg) -- pause if pausek lines have been printed.  'msg' is
               message, return 1st char of response, leaves others"
(let ((c nil))
  (if msg (if (< pausec pausek) (incf pausec)
	    (progn (format T "more - ~A > " msg)
		   (clear-input) (setq c (read-char))
		   (setq pausec 0)))
      (setq pausec 0))
 c)
)

(defun sharing ()
"sharing  -- returns number of active nodes for which refc > 1."
  (*when (>!! REFC (!! 1)) (*sum (!! 1)))
)

(defun pre (x)
"(pre formula) -- preprocesses formula for equations, associative ops"
 (cond ((atom x) x)
       ((eq (car x) '=) (equate (list-pre (cdr x))))
       ((member (car x) '(+ * and or)) (bal (list-pre x)))
       ((eq (car x) 'list) (form-list (cdr x) '[]))
       ((eq (car x) 'list*) (form-list (cdr x) nil))
       (T (list-pre x)))
)

; fin=[] - list ends with [];  fin=nil - list ends with last entry

(defun form-list (lst fin)
"(form-list list fin) -- returns rep of list ending according to fin"
 (cond ((null lst) fin)
       ((null (cdr lst)) (if fin (list 'cons (car lst) fin) (car lst)))
       (T (list 'cons (car lst) (form-list (cdr lst) fin))))
)

(defun list-pre (x) (mapcar #'pre x))

(defun equate (L)
"(equate L) - returns equations for forms in L"
 (if (<= (length L) 2)
     (cons '= L)
     (list 'and (list '= (first l) (second l)) (equate (cdr l))))
)

(defun bal (frm)
"(bal frm) -- returns balanced tree rep of associative form"
 (prog (fl o a s1 s2 m)
  (if (atom frm) (return frm))
  (setq o (car frm)) (setq a (cdr frm)) (setq fl (length a))
  (if (= fl 1) (return (car a)))
  (if (= fl 2) (return frm))
  (setq m (ash fl -1))
  (multiple-value-setq (s1 s2) (split a m))
  (return (list o (bal (cons o s1)) (bal (cons o s2))))
 ))

(defun split (L m)
"(split L m) -- returns (list of first m entries of L, rest of L)"
 (let (r)
  (labels ((sp (l m)
	       (if (zerop m) (progn (setq r l) nil)
		   (cons (car l) (sp (cdr l) (1- m))) )))
     (values (sp L m) r)))
)
